from tempfile import mkdtemp
from multiprocessing import freeze_support
from inspect import getsource

from kivy.clock import Clock
from kivy.lang.builder import Builder
from kivy.properties import BooleanProperty, NumericProperty

import networkx as nx
from networkx import grid_2d_graph

from ELiDE.game import GameApp, GameScreen, GridBoard


def make_grid() -> nx.Graph:



def game_start(engine: "LiSE.Engine") -> None:

